##
##
##   ███████ ███    ███  ██████     ███    ███ ██    ██ ██      ████████ ██        ██████  █████  ███    ███ ███████ ██████   █████
##   ██      ████  ████ ██          ████  ████ ██    ██ ██         ██    ██       ██      ██   ██ ████  ████ ██      ██   ██ ██   ██
##   █████   ██ ████ ██ ██          ██ ████ ██ ██    ██ ██         ██    ██ █████ ██      ███████ ██ ████ ██ █████   ██████  ███████
##   ██      ██  ██  ██ ██          ██  ██  ██ ██    ██ ██         ██    ██       ██      ██   ██ ██  ██  ██ ██      ██   ██ ██   ██
##   ██      ██      ██  ██████     ██      ██  ██████  ███████    ██    ██        ██████ ██   ██ ██      ██ ███████ ██   ██ ██   ██
##
##
# Font - ANSI Regular - https://patorjk.com/software/taag/#p=display&f=ANSI%20Regular&t=Play%20Skeleton%20Animation


from rich.repr import T
from FMC_Camera import FMC_Camera


import threading
import pathos.multiprocessing as pathos_mp
from pathos.helpers import mp as pathos_mp_helper
import time
import platform
import datetime

import cv2
import matplotlib.pyplot as plt
import numpy as np
from pathlib import Path

from rich import print
from rich import inspect
from rich.progress import track
from rich.traceback import install

install(show_locals=True)
from rich.console import Console

rich_console = Console()
from rich.progress import track


class FMC_MultiCamera:
    ##
    ##
    ##                  ██ ███    ██ ██ ████████
    ##                  ██ ████   ██ ██    ██
    ##                  ██ ██ ██  ██ ██    ██
    ##                  ██ ██  ██ ██ ██    ██
    ##  ███████ ███████ ██ ██   ████ ██    ██    ███████ ███████
    ##
    ##
    ##
    def __init__(
        self,
        rec_name=None,
        freemocap_data_folder=None,
        num_cams=0,
        cams_to_use_list=None,
        rotation_codes_list=None,
        show_multi_cam_stream_bool=True,
        save_multi_cam_to_mp4=False,
        save_each_cam_to_mp4=True,
        save_to_h5=False,
        save_log_file=False,
    ):

        self._init_start_time = time.time_ns()  # the precision is aspirational, lol

        if rec_name is None:
            self._rec_name = datetime.datetime.now().strftime(
                "sesh_%Y-%m-%d_%H_%M_%S_mc"
            )
        else:
            self._rec_name = rec_name

        if freemocap_data_folder is None:
            self._save_path = Path("data/" + self._rec_name)
        else:
            self._save_path = Path(freemocap_data_folder) / self._rec_name

        if (
            freemocap_data_folder
            or save_multi_cam_to_mp4
            or save_to_h5
            or save_log_file
            or save_each_cam_to_mp4
        ):
            self._save_path.mkdir(parents=True, exist_ok=True)
            self._path_to_log_file = Path(
                str(self._save_path / self._rec_name) + "_log.txt"
            )

        self._save_multi_cam_to_mp4 = save_multi_cam_to_mp4
        if self._save_multi_cam_to_mp4:
            self._output_multi_cam_video_object = (
                None  # build this after we get the first multi_frame_image
            )

        self._save_each_cam_to_mp4 = save_each_cam_to_mp4
        if self._save_each_cam_to_mp4:
            self._each_cam_video_writer_object_list = (
                None  # build this after we get the first multi_frame_image
            )

        # log_file_obj = open(self._path_to_log_file,'w+')
        # rich_console = Console(file=log_file_obj)
        rich_console = Console()

        if cams_to_use_list:
            self._num_cams = len(cams_to_use_list)
        else:
            self._num_cams = num_cams
        self._cams_to_use_list = cams_to_use_list

        self._rotation_codes_list = rotation_codes_list

        self._show_multi_cam_stream_bool = show_multi_cam_stream_bool

        self.num_frames = 0

        rich_console.rule("Starting FreeMoCap MultiCam!")

    ###
    ###
    ###   ██████  ██████   ██████  ██████  ███████ ██████  ████████ ██ ███████ ███████
    ###   ██   ██ ██   ██ ██    ██ ██   ██ ██      ██   ██    ██    ██ ██      ██
    ###   ██████  ██████  ██    ██ ██████  █████   ██████     ██    ██ █████   ███████
    ###   ██      ██   ██ ██    ██ ██      ██      ██   ██    ██    ██ ██           ██
    ###   ██      ██   ██  ██████  ██      ███████ ██   ██    ██    ██ ███████ ███████
    ###
    ###

    @property
    def num_cams(self):
        """How many cameras to use"""
        return self._num_cams

    @num_cams.setter
    def num_cams(self, num_cams_input):
        """Define how many cameras to use"""
        self._num_cams = num_cams_input

    @property
    def cams_to_use_list(self):
        """a list denoting which camera IDs to use.
        If both this and `num_cams` are set to default value (`0` and `None`, respectively), use all available cameras).
        """
        return self._cams_to_use_list

    @property
    def init_start_time(self):
        """Time that this object was initialized (UNIX Epoch time from `time.time_ns())

        Returns:
            float: time this object was created (UNIX Epoch time from `time.time_ns())
        """
        return self._init_start_time

    @property
    def each_cam_timestamps_unix_ns(self):  # not really a property though?
        return self._each_cam_timestamps_unix_ns

    ###
    ###
    ###   ███    ███ ███████ ████████ ██   ██  ██████  ██████  ███████
    ###   ████  ████ ██         ██    ██   ██ ██    ██ ██   ██ ██
    ###   ██ ████ ██ █████      ██    ███████ ██    ██ ██   ██ ███████
    ###   ██  ██  ██ ██         ██    ██   ██ ██    ██ ██   ██      ██
    ###   ██      ██ ███████    ██    ██   ██  ██████  ██████  ███████

    #################################################################
    ###
    ###
    ###   ███████ ████████  █████  ██████  ████████
    ###   ██         ██    ██   ██ ██   ██    ██
    ###   ███████    ██    ███████ ██████     ██
    ###        ██    ██    ██   ██ ██   ██    ██
    ###   ███████    ██    ██   ██ ██   ██    ██
    ###
    ###
    ###
    ###
    def start(self, standalone_mode=False):
        self.standalone_mode = standalone_mode
        if self.cams_to_use_list is None:
            self.find_available_cameras()

        # pathos_pool = ProcessPool(nodes=1)
        # pathos_pool.map(self.start_multi_cam_thread_pool)
        # self.start_multi_cam_thread_pool() #threads are mostly fine, but I think processes will be better eventually?
        self.start_multi_cam_process_pool()
        if standalone_mode:
            self.create_diagnostic_images()

    ##
    ##
    ##    ███████ ██ ███    ██ ██████       ██████  █████  ███    ███ ███████
    ##    ██      ██ ████   ██ ██   ██     ██      ██   ██ ████  ████ ██
    ##    █████   ██ ██ ██  ██ ██   ██     ██      ███████ ██ ████ ██ ███████
    ##    ██      ██ ██  ██ ██ ██   ██     ██      ██   ██ ██  ██  ██      ██
    ##    ██      ██ ██   ████ ██████       ██████ ██   ██ ██      ██ ███████
    ##
    ##
    ##

    def find_available_cameras(self):
        """find available webcams and return IDs in list."""

        if self.num_cams == 0:
            num_cams_to_check = 20  # man, i HOPE you're out there trying to record from more than 20 cameras!!!
        else:
            num_cams_to_check = self.num_cams

        if platform.system() == "Windows":
            capBackend = cv2.CAP_DSHOW
        else:
            capBackend = cv2.CAP_ANY

        self._cams_to_use_list = []
        for camNum in range(
            num_cams_to_check
        ):  # track(range(num_cams_to_check), description='[green]Finding available cameras...' ):
            # print('ATTEMPTING TO OPEN CAMERA AT PORT# {}'.format(camNum))
            cap = cv2.VideoCapture(camNum, capBackend)
            success, image = cap.read()
            if (
                success
            ):  # this is apparently a more power check than `cap.isOpened()` which sometimes returns false positives for unknown reason. Will lose the first frame if this method is ever applied to reading in videos instead of streaming from cameras
                # rich_console.print('SUCCESS - CAMERA FOUND at # {}'.format(camNum), style="bold cyan")
                self._cams_to_use_list.append(camNum)
                cap.release()
        else:
            pass
            # rich_console.print('No CAMERA FOUND at # {}'.format(camNum), style="magenta")

        rich_console.print("Found cameras at ", self._cams_to_use_list)

        self.num_cams = len(self._cams_to_use_list)
        return self.cams_to_use_list

    ###
    ###
    ###   ███████ ████████  █████  ██████  ████████     ██████  ██████   ██████   ██████ ███████ ███████ ███████     ██████   ██████   ██████  ██
    ###   ██         ██    ██   ██ ██   ██    ██        ██   ██ ██   ██ ██    ██ ██      ██      ██      ██          ██   ██ ██    ██ ██    ██ ██
    ###   ███████    ██    ███████ ██████     ██        ██████  ██████  ██    ██ ██      █████   ███████ ███████     ██████  ██    ██ ██    ██ ██
    ###        ██    ██    ██   ██ ██   ██    ██        ██      ██   ██ ██    ██ ██      ██           ██      ██     ██      ██    ██ ██    ██ ██
    ###   ███████    ██    ██   ██ ██   ██    ██        ██      ██   ██  ██████   ██████ ███████ ███████ ███████     ██       ██████   ██████  ███████
    ###
    ###
    ###
    def start_multi_cam_process_pool(self):
        """
        reates a ProcessPoolExecutor that creates an `FMC_Camera` object for each camera in `self.cams_to_use_list` and sets each to trigger `run_in_process`  mode (which just passes to `run_in_thread` mode)

        """
        mrManager = pathos_mp_helper.Manager()
        self.cam_frame_queue = mrManager.Queue(maxsize=self.num_cams)
        self.multi_cam_tuple_queue = mrManager.Queue()
        self.barrier = mrManager.Barrier(
            self.num_cams + 1
        )  # each cam + one more for the frame grabber
        self.exit_event = mrManager.Event()

        self._each_cam_timestamps_unix_ns = None
        self._cam_process = [None] * self.num_cams

        rich_console.rule("starting multi_cam_process_pool")

        in_frame_q_list = [self.cam_frame_queue] * self.num_cams
        in_barrier_list = [self.barrier] * self.num_cams
        in_exit_event_list = [self.exit_event] * self.num_cams

        if self._rotation_codes_list is None:
            self._rotation_codes_list = [None] * self.num_cams

        num_jobs = self.num_cams
        with pathos_mp.ProcessingPool(num_jobs) as self.cam_process_pool:
            self.cam_process_pool.amap(
                FMC_Camera,
                tuple(self.cams_to_use_list),
                tuple(self._rotation_codes_list),
                tuple(in_frame_q_list),
                tuple(in_barrier_list),
                tuple(in_exit_event_list),
            )
            self.incoming_tuple_grabber_thread = threading.Thread(
                target=self.grab_incoming_cam_tuples,
                name="Incoming Frame Tuple Grabber Thread",
            )
            self.incoming_tuple_grabber_thread.start()
            if self._show_multi_cam_stream_bool or self.standalone_mode:
                self.show_multi_cam_opencv()

    ###
    ###
    ###   ██████  ██████   █████  ██████      ██ ███    ██  ██████  ██████  ███    ███ ██ ███    ██  ██████      ████████ ██    ██ ██████  ██      ███████ ███████
    ###  ██       ██   ██ ██   ██ ██   ██     ██ ████   ██ ██      ██    ██ ████  ████ ██ ████   ██ ██              ██    ██    ██ ██   ██ ██      ██      ██
    ###  ██   ███ ██████  ███████ ██████      ██ ██ ██  ██ ██      ██    ██ ██ ████ ██ ██ ██ ██  ██ ██   ███        ██    ██    ██ ██████  ██      █████   ███████
    ###  ██    ██ ██   ██ ██   ██ ██   ██     ██ ██  ██ ██ ██      ██    ██ ██  ██  ██ ██ ██  ██ ██ ██    ██        ██    ██    ██ ██      ██      ██           ██
    ###   ██████  ██   ██ ██   ██ ██████      ██ ██   ████  ██████  ██████  ██      ██ ██ ██   ████  ██████         ██     ██████  ██      ███████ ███████ ███████
    ###
    ###

    def grab_incoming_cam_tuples(self):
        """grab incoming cam_tuples, put them into a multi_cam_tuple and stuff that into the multi_cam_tuple_queue"""

        # self.show_multi_cam_cv2()

        while not self.exit_event.is_set():
            self.barrier.wait()  # wait until each camera has grabbed an image and tagged their `barrier.wait()` signals

            self.num_frames += 1

            these_images_list = [None] * self.num_cams  # empty list of size (numCam)
            this_multi_cam_tuple_as_a_list = [None] * self.num_cams
            these_timestamps = np.ndarray(self.num_cams)

            for cam_num in range(self.num_cams):
                this_cam_image_timestamp_tuple = self.cam_frame_queue.get()
                this_cam_num = this_cam_image_timestamp_tuple[0]
                this_cam_index = self.cams_to_use_list.port_number(this_cam_num)
                these_images_list[this_cam_index] = this_cam_image_timestamp_tuple[1]
                these_timestamps[this_cam_index] = this_cam_image_timestamp_tuple[2]
                this_multi_cam_tuple_as_a_list[
                    this_cam_index
                ] = this_cam_image_timestamp_tuple

            this_multi_cam_tuple = tuple(this_multi_cam_tuple_as_a_list)

            if self._each_cam_timestamps_unix_ns is None:
                self._each_cam_timestamps_unix_ns = these_timestamps
            else:
                self._each_cam_timestamps_unix_ns = np.vstack(
                    (self._each_cam_timestamps_unix_ns, these_timestamps)
                )  # result will be numpy array with `num_frames` rows and `num_cams` columns

            self.multi_cam_tuple_queue.put(
                this_multi_cam_tuple, np.mean(these_timestamps)
            )

            rich_console.log(
                "Created a multi_cam_tuple - queue size: {}".format(
                    self.multi_cam_tuple_queue.qsize()
                )
            )

    def stitch_multicam_image(self, multi_cam_tuple):
        """Take in a multi-cam-tuple (containing images from each camera during the period of time defined by the attached timestamps)

        Args:
            multi_cam_tuple (tuple of tuples): The multi_cam_tuple produced by `self.grab_incoming_cam_tuples()` comprising a timestamp and a tuple from each of the cameras. Each camera's tuple consists of the camera's ID/Number, the image frame, and the timestamp that frame was recorded. The multi-camtimestamp is the mean of each of the timestamps from the individual cameras.
        Returns:
            multi_cam_image: an image of size camera_res_width*num_cams x camera_res_height The images are synchronized within the precision defined by the timestamps included in the individual camera tuples.
        """

        these_images_list = [None] * self.num_cams  # empty list of size (numCam)

        for this_cam_num in range(self.num_cams):
            these_images_list[this_cam_num] = multi_cam_tuple[this_cam_num][
                1
            ]  # and that's how you navigate nested tuples, lol

        multi_cam_image = np.hstack(
            these_images_list
        )  # create multiFrame_image by stitching together (horizontally stacking) incoming camera images (matrices)
        return multi_cam_image

        # rich_console.log('Created a multi_cam_image - queue size: {}'.format(self.multi_cam_tuple_queue.qsize()))

    ###
    ###
    ###  ███████ ██   ██  ██████  ██     ██     ██    ██ ██ ██████
    ###  ██      ██   ██ ██    ██ ██     ██     ██    ██ ██ ██   ██
    ###  ███████ ███████ ██    ██ ██  █  ██     ██    ██ ██ ██   ██
    ###       ██ ██   ██ ██    ██ ██ ███ ██      ██  ██  ██ ██   ██
    ###  ███████ ██   ██  ██████   ███ ███        ████   ██ ██████
    ###
    ###                                                             6

    def show_multi_cam_opencv(self):
        """display multi_cam_image using `cv2.imshow` and maybe save it to mp4, who knows?"""
        rich_console.rule("Launching Multi Cam Viewer")
        while not self.exit_event.is_set():
            if not self.multi_cam_tuple_queue.empty():

                this_multi_cam_tuple = self.multi_cam_tuple_queue.get()
                self.save_synchronized_videos(this_multi_cam_tuple)
                this_multi_cam_image = self.stitch_multicam_image(this_multi_cam_tuple)

                if self._save_multi_cam_to_mp4:
                    if self._output_multi_cam_video_object is None:
                        self.initialize_multi_cam_output_video(this_multi_cam_image)

                    self._output_multi_cam_video_object.write(this_multi_cam_image)

                if self._save_each_cam_to_mp4:
                    if self._each_cam_video_writer_object_list is None:
                        self.initialize_each_cam_output_video(this_multi_cam_tuple)

                    for this_cam_num in range(self.num_cams):
                        self._each_cam_video_writer_object_list[this_cam_num].write(
                            this_multi_cam_tuple[this_cam_num][1]
                        )

                cv2.imshow(self._rec_name, this_multi_cam_image)
                key = cv2.waitKey(1)

                if key == 27:  # exit on ESC
                    break

                if (
                    cv2.getWindowProperty(self._rec_name, cv2.WND_PROP_VISIBLE) < 1
                ):  # break loop if window closed
                    break

        cv2.destroyAllWindows()

        if self._save_multi_cam_to_mp4:
            self._output_multi_cam_video_object.release()

        if self._save_each_cam_to_mp4:
            for this_cam_num in range(self.num_cams):
                self._each_cam_video_writer_object_list[this_cam_num].release()

        rich_console.rule("Shutting down MultiCamera Viewer")
        self.exit_event.set()  # send the 'Exit' signal to everyone.

    def save_synchronized_videos(self, multi_cam_tuple):
        """save camera streams into individual videos (that can be processed with pre-alpha freemocap"""
        these_images_list = [None] * self.num_cams  # empty list of size (numCam)

        for this_cam_num in range(self.num_cams):
            this_cam_image = multi_cam_tuple[this_cam_num][
                1
            ]  # and that's how you navigate nested tuples, lol

    ###
    ###
    ###  ██ ███    ██ ██ ████████      ██████  ██    ██ ████████     ██    ██ ██ ██████
    ###  ██ ████   ██ ██    ██        ██    ██ ██    ██    ██        ██    ██ ██ ██   ██
    ###  ██ ██ ██  ██ ██    ██        ██    ██ ██    ██    ██        ██    ██ ██ ██   ██
    ###  ██ ██  ██ ██ ██    ██        ██    ██ ██    ██    ██         ██  ██  ██ ██   ██
    ###  ██ ██   ████ ██    ██         ██████   ██████     ██          ████   ██ ██████
    ###
    ###

    def initialize_multi_cam_output_video(self, multi_cam_image):
        """use  the multi_cam_image to initialize the video
            multicam image will be of size  [cam_resolution_width*num_cams by cam_resolution_height]
        Args:
            multi_cam_image ([type]): [description]
        """
        # create video save object
        (
            self._multi_cam_image_height,
            self._multi_cam_image_width,
            channels,
        ) = multi_cam_image.shape

        self.multi_cam_image_size = (
            self._multi_cam_image_width,
            self._multi_cam_image_height,
        )
        fourcc = cv2.VideoWriter_fourcc(*"DIVX")
        self._output_multi_cam_vid_fileName = (
            str(self._save_path / self._rec_name) + "_outVid.mp4"
        )
        fps = 25  # this is a bad and stupid guess, but it's actually kinda tricky to guess what is the right thing to put here. I'll fix this later and expect videos to be a bit faster than usual :(
        self._output_multi_cam_video_object = cv2.VideoWriter(
            self._output_multi_cam_vid_fileName, fourcc, fps, self.multi_cam_image_size
        )

    def initialize_each_cam_output_video(self, multi_cam_tuple):

        self._each_cam_video_writer_object_list = [None] * self.num_cams
        self._each_cam_video_filename_list = [None] * self.num_cams

        self.syncedVidsPath = self._save_path / "Synchronized_Videos"
        self.syncedVidsPath.mkdir(parents=True)

        for this_cam_num in range(self.num_cams):
            # create video save object
            this_cam_image = multi_cam_tuple[this_cam_num][1]
            this_cam_timestamp_unix_ns = multi_cam_tuple[this_cam_num][2]
            (
                self._cam_image_height,
                self._cam_image_width,
                channels,
            ) = this_cam_image.shape

            self._cam_image_size = (self._cam_image_width, self._cam_image_height)
            fourcc = cv2.VideoWriter_fourcc(*"DIVX")

            this_cam_vid_filename = (
                self._rec_name + "_Camera_" + str(this_cam_num) + "_synchronized.mp4"
            )
            self._each_cam_video_filename_list[this_cam_num] = (
                self.syncedVidsPath / this_cam_vid_filename
            )

            fps = 25  # this is a bad and stupid guess, but it's actually kinda tricky to guess what is the right thing to put here. I'll fix this later and expect videos to be a bit faster than usual :(
            self._each_cam_video_writer_object_list[this_cam_num] = cv2.VideoWriter(
                str(self._each_cam_video_filename_list[this_cam_num]),
                fourcc,
                fps,
                self._cam_image_size,
            )

    ##
    ##
    ##
    ##   ██████  ██       ██████  ████████     ████████ ██ ███    ███ ███████ ███████ ████████  █████  ███    ███ ██████  ███████
    ##   ██   ██ ██      ██    ██    ██           ██    ██ ████  ████ ██      ██         ██    ██   ██ ████  ████ ██   ██ ██
    ##   ██████  ██      ██    ██    ██           ██    ██ ██ ████ ██ █████   ███████    ██    ███████ ██ ████ ██ ██████  ███████
    ##   ██      ██      ██    ██    ██           ██    ██ ██  ██  ██ ██           ██    ██    ██   ██ ██  ██  ██ ██           ██
    ##   ██      ███████  ██████     ██           ██    ██ ██      ██ ███████ ███████    ██    ██   ██ ██      ██ ██      ███████
    ##
    ##

    def create_diagnostic_images(self):
        """plot some diagnostics to assess quality of camera sync"""
        try:
            camTimestamps = (
                self.each_cam_timestamps_unix_ns - self.init_start_time
            ) / 1e9  # subtract start time and convert to sec

            npy_save_path = self._save_path / "multi_cam_timestamps_frameNum_camNum.npy"
            np.save(str(npy_save_path), camTimestamps)
            rich_console.log("saving timestamp data to - " + str(npy_save_path))

            meanMultiFrameTimestamp = np.mean(camTimestamps, axis=1)
            meanMultiFrameTimespan = np.max(camTimestamps, axis=1) - np.min(
                camTimestamps, axis=1
            )  # what was the timespan covered by each frame
            plt.ion()
            fig = plt.figure(figsize=(18, 10))
            max_frame_duration = 0.1
            ax1 = plt.subplot(
                231,
                title="Camera Frame Timestamp vs Frame#",
                xlabel="Frame#",
                ylabel="Timestamp (sec)",
            )
            ax2 = plt.subplot(
                232,
                ylim=(0, max_frame_duration),
                title="Camera Frame Duration Trace",
                xlabel="Frame#",
                ylabel="Duration (sec)",
            )
            ax3 = plt.subplot(
                233,
                xlim=(0, max_frame_duration),
                title="Camera Frame Duration Histogram (count)",
                xlabel="Duration(s, 1ms bins)",
                ylabel="Probability",
            )
            ax4 = plt.subplot(
                234,
                title="MuliFrame Timestamp vs Frame#",
                xlabel="Frame#",
                ylabel="Timestamp (sec)",
            )
            ax5 = plt.subplot(
                235,
                ylim=(0, max_frame_duration),
                title="Multi Frame Duration/Span Trace",
                xlabel="Frame#",
                ylabel="Duration (sec)",
            )
            ax6 = plt.subplot(
                236,
                xlim=(0, max_frame_duration),
                title="MultiFrame Duration Histogram (count)",
                xlabel="Duration(s, 1ms bins)",
                ylabel="Probability",
            )

            for camNum in range(self.num_cams):
                thisCamTimestamps = camTimestamps[:, camNum]
                ax1.plot(thisCamTimestamps, label="Camera#" + str(camNum))
                ax1.legend()
                ax2.plot(np.diff(thisCamTimestamps), ".")
                ax3.hist(
                    np.diff(thisCamTimestamps),
                    bins=np.arange(0, max_frame_duration, 0.001),
                    alpha=0.5,
                )

            ax4.plot(
                meanMultiFrameTimestamp,
                color="darkslategrey",
                label="MultiFrame" + str(camNum),
            )
            ax5.plot(
                np.diff(meanMultiFrameTimestamp),
                ".",
                color="darkslategrey",
                label="Frame Duration",
            )
            ax5.plot(
                meanMultiFrameTimespan, ".", color="orangered", label="Frame TimeSpan"
            )
            ax5.legend()
            ax6.hist(
                np.diff(meanMultiFrameTimestamp),
                bins=np.arange(0, max_frame_duration, 0.001),
                density=True,
                alpha=0.5,
                color="darkslategrey",
                label="Frame Duration",
            )
            ax6.hist(
                np.diff(meanMultiFrameTimespan),
                bins=np.arange(0, max_frame_duration, 0.001),
                density=True,
                alpha=0.5,
                color="orangered",
                label="Frame Timespan",
            )
            ax5.legend()

            fig_save_path = self._save_path / "recording_diagnostics.png"
            plt.savefig(str(fig_save_path))
            rich_console.log("Saving diagnostic figure to - " + str(fig_save_path))
            plt.show()
            plt.waitforbuttonpress()
        except:
            rich_console.print_exception()
        f = 9

    ###
    ### ███████ ████████  ██████
    ### ██         ██    ██
    ### █████      ██    ██
    ### ██         ██    ██
    ### ███████    ██     ██████     ██ ██ ██
    ###

    def __enter__(self):
        """Context manager -  No need to do anything special on start"""
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Context manager -  on exit, set `self.exit_event` to activate shutdown sequence"""
        self.create_diagnostic_images()
        return self


###
###
###   ██ ███████                     ███    ███  █████  ██ ███    ██
###   ██ ██                          ████  ████ ██   ██ ██ ████   ██
###   ██ █████                       ██ ████ ██ ███████ ██ ██ ██  ██
###   ██ ██                          ██  ██  ██ ██   ██ ██ ██  ██ ██
###   ██ ██          ███████ ███████ ██      ██ ██   ██ ██ ██   ████ ███████ ███████
###
###


if __name__ == "__main__":
    pathos_mp_helper.freeze_support()
    console = Console()  # create rich console to catch and print exceptions

    import socket

    this_computer_name = socket.gethostname()

    freemocap_data_path = None
    in_rotation_codes_list = None
    in_cams_to_use_list = None

    if this_computer_name == "jon-hallway-XPS-8930":
        freemocap_data_path = Path("/home/jon/Dropbox/FreeMoCapProject/FreeMocap_Data")

    elif this_computer_name == "DESKTOP-DCG6K4F":
        freemocap_data_path = Path(
            r"C:\Users\jonma\Dropbox\FreeMoCapProject\FreeMocap_Data"
        )
        in_rotation_codes_list = [
            "cv2.ROTATE_90_COUNTERCLOCKWISE",
            "cv2.ROTATE_90_COUNTERCLOCKWISE",
            "cv2.ROTATE_90_CLOCKWISE",
            "cv2.ROTATE_90_CLOCKWISE",
            "cv2.ROTATE_90_CLOCKWISE",
        ]
        in_cams_to_use_list = [0, 1, 2, 3, 4]
    elif this_computer_name == "DESKTOP-V3D343U":
        freemocap_data_path = Path(
            r"C:\Users\WindowsPC_Hallway\Dropbox\FreeMoCapProject\FreeMocap_Data"
        )
    elif this_computer_name == "Jons-MacBook-Pro.local":
        freemocap_data_path = Path("/Users/jon/Dropbox/FreeMoCapProject")

    try:
        multi_cam = FMC_MultiCamera(
            freemocap_data_folder=str(freemocap_data_path),
            save_each_cam_to_mp4=True,
            rotation_codes_list=in_rotation_codes_list,
            cams_to_use_list=in_cams_to_use_list,
        )
        multi_cam.start(standalone_mode=True)

    except Exception:
        console.print_exception()



import tkinter as tk
from tkinter import Tk, Label, Button, Frame, Listbox, Entry, filedialog
from pathlib import Path

from numpy.lib.npyio import save

class setDataPathGUI:
    def __init__(self, master,session):
        self.master = master
        self.session = session
    
        chooseText = "Please choose the directory your FreeMoCap_data folder is located in"
        chooseLabel = Label(master, text=chooseText)
        chooseLabel.pack(side="top")

        self.currentPath = session.preferences['saved']['path_to_save'] 
        text = 'Current file path: ' + str(self.currentPath)
        self.currentPathText = tk.StringVar()
        self.currentPathText.set(text)
        currentPathLabel = Label(master,textvariable = self.currentPathText)
        currentPathLabel.pack()

        changepath_button = Button(master,text = 'Change Folder Path',command = self.openFileDialog)
        changepath_button.pack(side = 'left')

        proceed_button = Button(text="Proceed", command=self.proceed)
        stop_button = Button(text="Quit", command=self.stop)
        proceed_button.pack()
        stop_button.pack()



    def openFileDialog(self):
        self.currentPath = filedialog.askdirectory(
        title='Open a file',
        initialdir= Path.cwd())
        self.currentPathText.set('Current file path: ' + self.currentPath)

    
    def stop(self):
        self.master.destroy()

    
    def proceed(self):
        self.continueToRecording = True
        self.dataPath = self.currentPath
        self.master.destroy()


class DLCConfigPathGUI:
    def __init__(self,master,session,saved_dlc_paths):
        #saved_dlc_paths = ['one','two','three']
        self.master = master
        self.dlc_paths = saved_dlc_paths.copy()
        self.session = session

        introText = "Here are your currently saved DLC config paths"
        introLabel = Label(master, text=introText)
        introLabel.pack(side="top")

        #savedPathText = self.write_to_string(saved_dlc_paths)
        self.savedPathVar = tk.StringVar()
        #savedPathVar.set(savedPathText)
        self.update_dlc_path_label(self.dlc_paths)
        savedPathLabel = Label(master,textvariable=self.savedPathVar)
        savedPathLabel.pack()

        addpath_button = Button(master,text = 'Add a file path',command = self.openFileDialog)
        addpath_button.pack(side='top')

        clearpath_button = Button(master,text = 'Clear file paths',command = self.clearPaths)
        clearpath_button.pack(side='top')

        savepath_button =  Button(master,text = 'Save file paths and Proceed',command = self.savePaths)
        savepath_button.pack(side = 'top')


    def update_dlc_path_label(self,saved_dlc_paths):
        dlc_string_list = '\n'.join(saved_dlc_paths)
        self.savedPathVar.set(dlc_string_list)
        return dlc_string_list

    def openFileDialog(self):
        chosen_path = filedialog.askopenfilename(
        title='Open a file',
        initialdir= Path.cwd(),
        filetypes=[('Yaml','*.yaml')])
        self.dlc_paths.append(chosen_path)
        self.update_dlc_path_label(self.dlc_paths)
        f = 2
    
    def clearPaths(self):
        self.dlc_paths = []
        self.update_dlc_path_label(self.dlc_paths)

    def savePaths(self):
        self.config_paths= self.dlc_paths
        self.master.destroy()


        f = 2   
    



class setBlenderPathGUI:
    def __init__(self, master,session):
        self.master = master
        self.session = session
    
        chooseText = "Choose your Blender executable"
        chooseLabel = Label(master, text=chooseText)
        chooseLabel.pack(side="top")

        self.currentPath = Path.cwd()
        text = 'Current file path: ' + str(self.currentPath)
        self.currentPathText = tk.StringVar()
        self.currentPathText.set(text)
        currentPathLabel = Label(master,textvariable = self.currentPathText)
        currentPathLabel.pack()

        changepath_button = Button(master,text = 'Change Folder Path',command = self.openFileDialog)
        changepath_button.pack(side = 'left')

        proceed_button = Button(text="Proceed", command=self.proceed)
        stop_button = Button(text="Quit", command=self.stop)
        proceed_button.pack()
        stop_button.pack()



    def openFileDialog(self):
        self.currentPath = filedialog.askopenfilename(
        title='Open your Blender executable',
        initialdir= Path.cwd())
        self.currentPathText.set('Current file path: ' + self.currentPath)

    
    def stop(self):
        self.master.destroy()

    
    def proceed(self):
        self.continueToRecording = True
        self.dataPath = self.currentPath
        self.master.destroy()


def RunChooseBlenderPathGUI(session):
    root = tk.Tk()
    chosenBlenderPath = setBlenderPathGUI(root, session)
    root.mainloop()

    return chosenBlenderPath.dataPath




def RunChooseDataPathGUI(session):
    root = tk.Tk()
    chosenDataPath = setDataPathGUI(root, session)
    root.mainloop()

    return Path(chosenDataPath.dataPath)


def RunChooseDLCPathGUI(session,saved_dlc_paths):
    root = tk.Tk()
    chosenDLCPaths = DLCConfigPathGUI(root,session,saved_dlc_paths)
    root.mainloop()

    return chosenDLCPaths.config_paths

    f = 2

  


def determine_shared_charuco_board_views(self):
    pass
    # # determine paired board views
    # for this_cam_num in range(multi_cam.num_cams):
    #     if full_charuco_detected_on_this_frame[this_cam_num]:
    #         for this_other_camera_num in range(multi_cam.num_cams):
    #             if this_other_camera_num != this_cam_num:
    #                 if full_charuco_detected_on_this_frame[this_other_camera_num]:
    #                     each_cameras_shared_board_view_count_array[this_cam_num,
    #                     this_other_camera_num] += 1
    # each_cameras_shared_board_view_count_total = np.sum(
    # each_cameras_shared_board_view_count_array, axis=1)
    # # print(each_cameras_shared_board_view_count)
    #
    # for this_pair_num, this_cam_pair in enumerate(camera_pair_list):
    #     camera_one_id = this_cam_pair[0]
    #     camera_two_id = this_cam_pair[1]
    #
    #     if full_charuco_detected_on_this_frame[camera_one_id] and
    #     full_charuco_detected_on_this_frame[
    #         camera_two_id]:
    #         camera_pair_joint_view_count[this_pair_num] += 1



import cv2
import numpy as np
from rich import print

from src.cameras.capture.dataclasses.frame_payload import FramePayload
from src.core_processor.camera_calibration.calibration_dataclasses import LensDistortionCalibrationData
from src.core_processor.camera_calibration.charuco_board_detection.board_detector import BoardDetector
from src.core_processor.camera_calibration.calibration_diagnostics_visualizer import CalibrationDiagnosticsVisualizer
from src.core_processor.camera_calibration.charuco_board_detection.dataclasses.charuco_view_data import CharucoViewData
from src.core_processor.camera_calibration.charuco_board_detection.charuco_image_annotator import \
    annotate_image_with_charuco_data


class LensDistortionCalibrator:
    def __init__(self,
                 board_detector: BoardDetector = BoardDetector(),
                 show_calibration_diagnostics_visualizer=False, ):
        self._board_detector = board_detector
        self._current_calibration = None
        self._all_charuco_views = []
        self._num_charuco_views = 0
        self._best_combo_of_charuco_views = []
        self._previous_best_calibrations = []
        self.show_calibration_diagnostics_visualizer = show_calibration_diagnostics_visualizer
        if self.show_calibration_diagnostics_visualizer:
            self.calibration_diagnostics_visualizer = CalibrationDiagnosticsVisualizer()

    def process_incoming_frame(self,
                               raw_frame_payload: FramePayload):

        charuco_frame_payload = self._board_detector.detect_charuco_board(raw_frame_payload)

        charuco_view_data = charuco_frame_payload.charuco_view_data

        if not charuco_view_data.full_board_found and self._current_calibration is None:
            return charuco_frame_payload.annotated_image

        if charuco_view_data.full_board_found:
            self._all_charuco_views.append(charuco_view_data)
            self._num_charuco_views = len(self._all_charuco_views)

            if self._num_charuco_views < 2:
                return charuco_frame_payload.annotated_image

            self.estimate_lens_distortion(charuco_view_data)  # this is where the magic happens

        if self._current_calibration is None:
            self._current_calibration = LensDistortionCalibrationData(charuco_view_data.image_width,
                                                                      charuco_view_data.image_height)
            return raw_frame_payload.image

        undistorted_image_for_debug = self.undistort_image_with_invalid_pixels_as_black(raw_frame_payload.image,
                                                                                        self._current_calibration)
        if charuco_view_data.some_charuco_corners_found:
            annotate_image_with_charuco_data(
                undistorted_image_for_debug,  # this image will have stuff drawn on top of it inside this function
                charuco_view_data,
                self._board_detector.number_of_charuco_corners
            )

        if self.show_calibration_diagnostics_visualizer:
            self.calibration_diagnostics_visualizer.update_image_subplot(undistorted_image_for_debug)

        undistorted_image = cv2.undistort(raw_frame_payload.image,
                                          self._current_calibration.camera_matrix,
                                          self._current_calibration.lens_distortion_coefficients,
                                          )

        return undistorted_image

    def estimate_lens_distortion(self, new_charuco_view: CharucoViewData):
        """
        adapted from - https://mecaruco2.readthedocs.io/en/latest/notebooks_rst/Aruco/sandbox/ludovic/aruco_calibration_rotation.html

        Calibrates the camera using the charuco data using a RANSAC-like method - randomly selects board views and saves the combo that minimized reprojection error

        helpful resources -
        https://docs.opencv.org/4.x/d9/d0c/group__calib3d.html <-opencv's 3d camera calibration docs
        https://www.adamsmith.haus/python/docs/cv2.aruco.calibrateCameraCharucoExtended
        https://www.mathworks.com/help/vision/ug/camera-calibration.html
        https://docs.opencv.org/3.4/d9/db7/tutorial_py_table_of_contents_calib3d.html
        https://calib.io/blogs/knowledge-base/camera-models <-a good one
        """
        print("CAMERA CALIBRATION")

        if self._current_calibration is None:
            current_best_reprojection_error = 1e9  # put a big fake number on the first frame
        else:
            current_best_reprojection_error = self._current_calibration.reprojection_error

        image_width = new_charuco_view.image_width
        image_height = new_charuco_view.image_height

        default_calibration = LensDistortionCalibrationData(image_width, image_height)

        num_charuco_views_per_combo = 5
        if self._num_charuco_views < num_charuco_views_per_combo:
            list_of_combos_of_charuco_views = [self._all_charuco_views]
        else:
            list_of_combos_of_charuco_views = self.generate_combos_of_charuco_views(new_charuco_view,
                                                                                    num_combos_to_generate=10,
                                                                                    num_views_per_combo=num_charuco_views_per_combo,
                                                                                    )

        # flag definitions -> https://docs.opencv.org/4.x/d9/d0c/group__calib3d.html#ga3207604e4b1a1758aa66acb6ed5aa65d
        flags = (cv2.CALIB_USE_INTRINSIC_GUESS +
                 cv2.CALIB_ZERO_TANGENT_DIST +
                 cv2.cv2.CALIB_FIX_ASPECT_RATIO +
                 cv2.CALIB_FIX_PRINCIPAL_POINT +
                 # cv2.CALIB_FIX_FOCAL_LENGTH
                 cv2.CALIB_RATIONAL_MODEL
                 # cv2.CALIB_THIN_PRISM_MODEL
                 # cv2.CALIB_TILTED_MODEL
                 )

        for this_combo_of_charuco_views in list_of_combos_of_charuco_views:
            for this_charuco_view in this_combo_of_charuco_views:
                # https://docs.opencv.org/4.x/d9/d6a/group__aruco.html#gacf03e5afb0bc516b73028cf209984a06
                (this_reprojection_error,
                 this_camera_matrix,
                 these_lens_distortion_coefficients,
                 these_rotation_vectors_of_the_board,
                 these_translation_vectors_of_the_board,
                 this_lens_distortion_std_dev,
                 this_camera_location_std_dev,
                 these_reprojection_error_per_view
                 ) = cv2.aruco.calibrateCameraCharucoExtended(
                    charucoCorners=[this_charuco_view.charuco_corners],
                    charucoIds=[this_charuco_view.charuco_ids],
                    board=self._board_detector.charuco_board,
                    imageSize=(image_width, image_height),
                    cameraMatrix=default_calibration.camera_matrix,
                    distCoeffs=default_calibration.lens_distortion_coefficients,
                    flags=flags,
                    criteria=(cv2.TERM_CRITERIA_EPS & cv2.TERM_CRITERIA_COUNT, 10000, 1e-9))

                if self.show_calibration_diagnostics_visualizer:
                    self.calibration_diagnostics_visualizer.update_reprojection_error_subplot(this_reprojection_error,
                                                                                              current_best_reprojection_error,
                                                                                              )
                    if self._current_calibration is not None:
                        self.calibration_diagnostics_visualizer.update_calibration_text_overlay(self._current_calibration)

                if this_reprojection_error < current_best_reprojection_error:

                    this_calibration = LensDistortionCalibrationData(
                        reprojection_error=this_reprojection_error,
                        camera_matrix=this_camera_matrix,
                        lens_distortion_coefficients=these_lens_distortion_coefficients,
                        image_width=image_width,
                        image_height=image_height,
                        rotation_vectors_of_the_board=these_rotation_vectors_of_the_board,
                        translation_vectors_of_the_board=these_translation_vectors_of_the_board,
                        lens_distortion_std_dev=this_lens_distortion_std_dev,
                        camera_location_std_dev=this_camera_location_std_dev)

                    if self.is_this_calibration_valid(this_calibration):
                        self._current_calibration = this_calibration
                        self._previous_best_calibrations.append(self._current_calibration)

    def generate_combos_of_charuco_views(self,
                                         new_charuco_view,
                                         num_combos_to_generate=10,
                                         num_views_per_combo=5):
        """
        return -
        1x current best combo plus new view
        num_combos_to_generate/2x random combo WITH new view appended
        num_combos_to_generate/2x random combo WITHOUT new view appended
        """

        list_of_combos_of_charuco_views = []

        # add default calibration (i.e. option not to change anything)
        # list_of_combos_of_charuco_views.append([self.get_default_calibration(new_charuco_view)])

        # add previous best view
        if self._best_combo_of_charuco_views:
            list_of_combos_of_charuco_views.append(self._best_combo_of_charuco_views.copy())

        # add previous best view + new view
        best_combo_plus_new = self._best_combo_of_charuco_views.copy()
        best_combo_plus_new.append(new_charuco_view)
        list_of_combos_of_charuco_views.append(best_combo_plus_new)

        # add all previous best calibrations
        # list_of_combos_of_charuco_views.append([cc for cc in self._previous_best_calibrations])

        # and a some random combos - half with the new view and half without it
        for this_iter in range(num_combos_to_generate):
            this_random_sample_of_charuco_views = np.random.choice(num_views_per_combo,
                                                                   size=min(len(self._all_charuco_views),
                                                                            num_views_per_combo),
                                                                   replace=False)

            this_combo = [self._all_charuco_views[cc] for cc in this_random_sample_of_charuco_views]

            if this_iter < np.ceil(num_views_per_combo / 2):
                this_combo.append(new_charuco_view)

            list_of_combos_of_charuco_views.append(this_combo)

        return list_of_combos_of_charuco_views

    def is_this_calibration_valid(self, this_calibration):

        is_valid = False
        if not self.lens_distortion_is_monotonic(this_calibration):  # not currently working - returns True every time
            return False

        if self.check_if_too_many_invalid_pixels(this_calibration):
            return False

        return True

    def lens_distortion_is_monotonic(self, this_calibration, assume_wide_angle_lens=True) -> bool:
        """
        check if the lens distortion coefficients are valid by checking it the resulting undistorted image is monotonically increasing along each diagonal (i.e. the image doesn't wrap around itself)

        per opencv calib3d docs, calibrations are valid if -

        1 + k1*r^2 + k2*r^r + k3*r6

        is monotonically increasing (pincushion distortion) or monotonically decreasing (barrel distortion).

        non-monotonic behavior means the calibration is invalid, so we check for that by examining the `np.diff` of a vector of points extending from image center to the top left coners to make sure it never goes negative(NOTE - check other directions?)

        since most cameras are wide-angle (i.e. they give a wider field of view than a pinhole camera), we may constrain things further be requiring things be monotonically increasing (i.e. only return pincushion distortions to correct for the barrel distortion of a wide-angle field of view)


        see -
        https://www.mathworks.com/help/vision/ug/camera-calibration.html <- matlab docs (best ones, maybe)
        https://docs.opencv.org/4.x/d9/d0c/group__calib3d.html <-mathematical defintion of lens distortion
        https://docs.opencv.org/4.x/d9/d0c/group__calib3d.html#ga3207604e4b1a1758aa66acb6ed5aa65d <- docs explaining lens_distortion_coefficients
        https://github.com/opencv/opencv/issues/15992 <-conversation about the fact that opencv doesn't check for this kinda thing

        """
        calibration_valid = False

        # following naming conventions of open cv docs
        k1 = this_calibration.lens_distortion_coefficients[0]
        k2 = this_calibration.lens_distortion_coefficients[1]
        p1 = this_calibration.lens_distortion_coefficients[2]  # I think p1 and p2 are for tangential distortion...
        p2 = this_calibration.lens_distortion_coefficients[
            3]  # ...which I might want to check on later? Currently fixed at Zero
        k3 = this_calibration.lens_distortion_coefficients[4]

        number_of_points = 40
        original_x = np.linspace(this_calibration.image_width, number_of_points)
        original_y = np.linspace(this_calibration.image_height, number_of_points)

        original_points_xx, original_points_yy = np.meshgrid(original_x, original_y)

        original_points_x = original_points_xx.flatten()  # transform 2D grid with 1D array
        original_points_y = original_points_yy.flatten()  # transform 2D grid with 1D array

        original_points_xy = np.vstack((original_points_x, original_points_y))

        distorted_points_xy = cv2.undistortPoints(original_points_xy,
                                                  this_calibration.camera_matrix,
                                                  this_calibration.lens_distortion_coefficients)

        distorted_points_x = np.squeeze(distorted_points_xy[:, :, 0])*this_calibration.image_width+this_calibration.image_width/2
        distorted_points_y = np.squeeze(distorted_points_xy[:, :, 1])*this_calibration.image_height+this_calibration.image_height/2

        if self.show_calibration_diagnostics_visualizer:
            self.calibration_diagnostics_visualizer.update_image_point_remapping_subplot(original_points_x,
                                                                                         original_points_y,
                                                                                         distorted_points_x,
                                                                                         distorted_points_y)
        return True

    def check_if_too_many_invalid_pixels(self, this_calibration, how_many_is_too_many=.1) -> bool:
        """
        create new camera matrix that will show full undistored image with black pixels in spots with no data.
        return FALSE if black pixels are greater than `how_many_is_too_many` proportion of the full image
        """

        dummy_image = np.ones((this_calibration.image_width, this_calibration.image_height), np.uint8) * 255

        undistorted_image = self.undistort_image_with_invalid_pixels_as_black(dummy_image, this_calibration)

        num_invalid_pixels = np.sum(undistorted_image == 0)
        proportion_invalid_pixels = num_invalid_pixels / dummy_image.size

        if proportion_invalid_pixels > how_many_is_too_many:
            return True
        return False

    @staticmethod
    def undistort_image_with_invalid_pixels_as_black(image, calibration):
        new_camera_matrix, valid_ROI_lbwh = cv2.getOptimalNewCameraMatrix(
            calibration.camera_matrix,
            calibration.lens_distortion_coefficients,
            (calibration.image_width, calibration.image_height),
            1,
            centerPrincipalPoint=True
        )

        # https://docs.opencv.org/4.5.5/d9/d0c/group__calib3d.html#ga69f2545a8b62a6b0fc2ee060dc30559d
        undistorted_image = cv2.undistort(image,
                                          calibration.camera_matrix,
                                          calibration.lens_distortion_coefficients,
                                          None,
                                          new_camera_matrix)

        return undistorted_image


if __name__ == "__main__":
    pass



import numpy as np
import pyqtgraph as pg
import pyqtgraph.opengl as gl
from pyqtgraph.dockarea.Dock import Dock
from pyqtgraph.dockarea.DockArea import DockArea
from pyqtgraph.Qt import QtWidgets

import cv2

from src.core_processor.camera_calibration.calibration_dataclasses import CameraCalibrationData



class CalibrationDiagnosticsVisualizer:
    def __init__(self):
        # dummy data to initialize subplots
        self.reprojection_error_all = np.empty(0)
        self.reprojection_error_current_best = np.empty(0)
        self.image_point_original = np.empty(0)
        self.image_point_remapped = np.empty(0)

        # create main app and window
        self.pyqtgraph_app = pg.mkQApp("DockArea Example")
        self.main_window_widget = QtWidgets.QMainWindow()
        self.dock_area = DockArea()
        self.main_window_widget.setCentralWidget(self.dock_area)
        self.main_window_widget.resize(1000, 500)
        self.main_window_widget.setWindowTitle('Lens Distortion Calibration Diagnostics (pyqtgraph)')

        # create widget/dock for image subplot (based on `GraphicsItems - ImageItem - video` example `python -m pyqtgraph.examples`)
        self.image_view_widget = pg.ImageView()

        self.undistorted_image_dock = Dock("Undistorted Image")
        self.undistorted_image_dock.addWidget(self.image_view_widget)
        self.dock_area.addDock(self.undistorted_image_dock)

        # create widget/dock for reprojection error subplot
        self.reprojection_error_plot_widget = pg.PlotWidget(title="Reprojection Error")
        self.reprojection_error_plot_widget.setLabel('left', "Reprojection Error", units='pixels')
        self.reprojection_error_plot_widget.setLabel('bottom', "Time", units='Frame#')
        self.reprojection_error_plot_widget.setYRange(0, 2)
        self.reprojection_error_plot_widget.addLegend()
        self.reprojection_error_all_line = self.reprojection_error_plot_widget.plot(self.reprojection_error_all,
                                                                                    pen=(255, 0, 128),
                                                                                    name="reprojection error- all")
        self.reprojection_error_current_best_line = self.reprojection_error_plot_widget.plot(
            self.reprojection_error_current_best,
            name="current best")

        self.reprojection_error_dock = Dock("Reprojection Error")
        self.reprojection_error_dock.addWidget(self.reprojection_error_plot_widget)
        self.dock_area.addDock(self.reprojection_error_dock, position='right')

        # create widget/dock lens distortion remapping subplot
        self.image_point_remapping_plot_widget = pg.PlotWidget(title="Image Point Remapping")
        self.image_point_remapping_plot_widget.setLabel('left', "Y-position", units='pixels')
        self.image_point_remapping_plot_widget.setLabel('bottom', "X-position", units='pixels')
        self.image_point_remapping_plot_widget.addLegend()

        self.image_point_original_dottos = self.image_point_remapping_plot_widget.plot(self.image_point_original,
                                                                                       self.image_point_original,
                                                                                       pen=None,
                                                                                       symbol='o',
                                                                                       symbolBrush=None,
                                                                                       symbolPen=(150, 150, 150),
                                                                                       name="image points - original")
        self.image_point_remapped_dottos = self.image_point_remapping_plot_widget.plot(self.image_point_remapped,
                                                                                       self.image_point_remapped,
                                                                                       symbol='+',
                                                                                       pen=None,
                                                                                       symbolBrush=None,
                                                                                       symbolPen=(50, 150, 210),
                                                                                       name="image points - remapped")
        self.image_point_remapping_plot_dock = Dock("Image Point Remapping")
        self.image_point_remapping_plot_dock.addWidget(self.image_point_remapping_plot_widget)
        self.dock_area.addDock(self.image_point_remapping_plot_dock)

        ###########
        # create widget/dock opengl 3d plot for charuco dottos
        self.opengl_3d_plot_widget = gl.GLViewWidget()
        self.opengl_grid_item = gl.GLGridItem()
        self.opengl_3d_plot_widget.addItem(self.opengl_grid_item)

        # create XYZ axes
        x_axis_line_array = np.array([[0, 0, 0], [1, 0, 0]])
        y_axis_line_array = np.array([[0, 0, 0], [0, 1, 0]])
        z_axis_line_array = np.array([[0, 0, 0], [0, 0, 1]])

        self.origin_x_axis_gl_lineplot_item = gl.GLLinePlotItem(pos=x_axis_line_array,
                                                                color=(1, 0, 0, 1),
                                                                width=1.,
                                                                antialias=True)
        self.origin_y_axis_gl_lineplot_item = gl.GLLinePlotItem(pos=y_axis_line_array,
                                                                color=(0, 1, 0, 1),
                                                                width=1.,
                                                                antialias=True)
        self.origin_z_axis_gl_lineplot_item = gl.GLLinePlotItem(pos=z_axis_line_array,
                                                                color=(0, 0, 1, 1),
                                                                width=1.,
                                                                antialias=True)
        self.opengl_3d_plot_widget.addItem(self.origin_x_axis_gl_lineplot_item)
        self.opengl_3d_plot_widget.addItem(self.origin_y_axis_gl_lineplot_item)
        self.opengl_3d_plot_widget.addItem(self.origin_z_axis_gl_lineplot_item)

        self.opengl_charuco_scatter_item = gl.GLScatterPlotItem(pos=(0, 0, 0), color=(1, 0, 1), size=1, pxMode=False)

        self.opengl_3d_plot_dock = Dock("Image Point Remapping")
        self.opengl_3d_plot_dock.addWidget(self.opengl_3d_plot_widget)
        self.dock_area.addDock(self.opengl_3d_plot_dock, 'right', self.image_point_remapping_plot_dock)

        ###
        # create widget/dock for calibration text window
        self.calibration_text_widget = QtWidgets.QLabel()

        self.calibration_text_dock = Dock("Calibration Data")
        self.calibration_text_dock.addWidget(self.calibration_text_widget)
        self.dock_area.addDock(self.calibration_text_dock, 'above', self.image_point_remapping_plot_dock)

        self._display_main_window()

    def _display_main_window(self):
        # display window
        self.main_window_widget.show_window()

    def update_image_subplot(self, new_image):
        self.image_view_widget.setImage(
            cv2.cvtColor(cv2.rotate(new_image, cv2.ROTATE_90_COUNTERCLOCKWISE), cv2.COLOR_BGR2RGB))

    def update_reprojection_error_subplot(self, reprojection_error_new, reprojection_error_current_best):
        self.reprojection_error_all = np.append(self.reprojection_error_all, reprojection_error_new)
        self.reprojection_error_current_best = np.append(self.reprojection_error_current_best,
                                                         reprojection_error_current_best)

        self.reprojection_error_all_line.setData(self.reprojection_error_all)
        self.reprojection_error_current_best_line.setData(self.reprojection_error_current_best)

    def update_calibration_text_overlay(self, calibration_data: CameraCalibrationData):
        lens_distortion_as_str = [f"{d[0]:.3f}" for d in calibration_data.lens_distortion_coefficients]
        calibration_text = f'reprojection error:{calibration_data.reprojection_error:.3f} \n {lens_distortion_as_str} \n {str(calibration_data.camera_matrix)}'
        self.calibration_text_widget.setText(calibration_text)

    def update_image_point_remapping_subplot(self, image_point_original_x, image_point_original_y,
                                             image_point_remapped_x, image_point_remapped_y):
        self.image_point_original_dottos.setData(x=image_point_original_x,
                                                 y=image_point_original_y)

        self.image_point_remapped_dottos.setData(x=image_point_remapped_x,
                                                 y=image_point_remapped_y)

    def update_3d_subplot(self, charuco_xyz: np.ndarray = None):
        if charuco_xyz is None:
            return
        self.opengl_charuco_scatter_item.setData(pos=charuco_xyz)

    def close(self):
        self.main_window_widget.close()



from typing import Dict

from singleton_decorator import singleton

from src.config.webcam_config import WebcamConfig


@singleton
class UserConfigService:
    _webcam_config: Dict[str, WebcamConfig] = {}

    @property
    def webcam_configs(self):
        return self._webcam_config

    def webcam_config_by_id(self, webcam_id: str):
        if webcam_id in self._webcam_config:
            return self._webcam_config[webcam_id]
        else:
            return WebcamConfig(webcam_id=int(webcam_id))

    def set_webcam_config(self, webcam_config: WebcamConfig):
        as_str = str(webcam_config.webcam_id)
        self._webcam_config[as_str] = webcam_config



from src.cameras.multicam_manager.cv_camera_manager import OpenCVCameraManager
from src.core_processor.mediapipe_skeleton_detection.mediapipe_skeleton_detection import (
    MediapipeSkeletonDetection,
)


class MediapipeSkeletonDetectionService:
    def __init__(self, cam_manager: OpenCVCameraManager):
        self._cam_manager = cam_manager

    def run(self, model_complexity: int):
        MediapipeSkeletonDetection(self._cam_manager).process(
            model_complexity=model_complexity
        )

    async def run_as_loop(self, webcam_id, cb, model_complexity: int):
        await MediapipeSkeletonDetection(self._cam_manager).process_as_frame_loop(
            webcam_id, model_complexity=model_complexity, cb=cb
        )



# -*- coding: utf-8 -*-
"""
built off the '3D Graphics- Scatter Plot' example from `python -m pyqtgraph.examples`
"""
from pathlib import Path
import sys

import pyqtgraph as pg
from pyqtgraph import functions as fn
from pyqtgraph.Qt import QtCore, QtGui
import pyqtgraph.opengl as gl
import numpy as np


class FMC_Visualizer:
    def __init__(self, fmc_session_obj = None, recording_session_path = None):

        #setup QT app window
        self.app = pg.mkQApp("FreeMoCap Data Visualizer")
        self.gl_view_widget = gl.GLViewWidget()
        self.gl_view_widget.show()
        self.gl_view_widget.setWindowTitle('FreeMoCap Data Visualizer')
        self.gl_view_widget.setCameraPosition(distance=10, elevation=-90, azimuth=-90)

        grid_distance = 10
        gl_x_plane_grid_item = gl.GLGridItem()        
        gl_x_plane_grid_item.rotate(90, 0, 1, 0)
        gl_x_plane_grid_item.translate(grid_distance, 0, 0)        
        self.gl_view_widget.addItem(gl_x_plane_grid_item)
        
        gl_y_plane_grid_item = gl.GLGridItem()
        gl_y_plane_grid_item.rotate(90, 1, 0, 0)
        gl_y_plane_grid_item.translate(0, grid_distance, 0)
        self.gl_view_widget.addItem(gl_y_plane_grid_item)
        
        gl_z_plane_grid_item = gl.GLGridItem()
        gl_z_plane_grid_item.translate(0, 0, grid_distance)
        self.gl_view_widget.addItem(gl_z_plane_grid_item)
        
        origin_location_array = np.array([0,0, 0])
        x_axis_line_array = np.array([[0,0, 0], [1,0,0]])
        y_axis_line_array = np.array([[0,0, 0], [0,1,0]])
        z_axis_line_array = np.array([[0,0, 0], [0,0,1]])
                        
        # self.origin_location_scatter_item = gl.GLScatterPlotItem(pos=origin_location_array, color=(1,1,1,1),  size=.1, pxMode=False)
        self.origin_x_axis_gl_lineplot_item = gl.GLLinePlotItem(pos=x_axis_line_array, color=(1,0,0,1), width=1., antialias=True)
        self.origin_y_axis_gl_lineplot_item = gl.GLLinePlotItem(pos=y_axis_line_array, color=(0,1,0,1), width=1., antialias=True)
        self.origin_z_axis_gl_lineplot_item = gl.GLLinePlotItem(pos=z_axis_line_array, color=(0,0,1,1), width=1., antialias=True)
        
        # self.gl_view_widget.addItem(self.origin_location_scatter_item)
        self.gl_view_widget.addItem(self.origin_x_axis_gl_lineplot_item)
        self.gl_view_widget.addItem(self.origin_y_axis_gl_lineplot_item)
        self.gl_view_widget.addItem(self.origin_z_axis_gl_lineplot_item)
        
        if fmc_session_obj is None:       
            if recording_session_path: 
                self.recording_session_path = Path(recording_session_path)
            else:
                #JSM NOTE - add something to load the last recorded session if nothing is specified
                pass
                
            # load charuco data
            self.charuco_3d_path = self.recording_session_path / "charuco_3d_points.npy"
            self.charuco_nFrames_nTrackedPoints_XYZ = np.load(str(self.charuco_3d_path))
            self.charuco_nFrames_nTrackedPoints_XYZ = self.charuco_nFrames_nTrackedPoints_XYZ/1000 #convert to meters

            # load mediapipe data
            self.mediapipe_3d_path = self.recording_session_path / "mediapipe_3d_points.npy"
            self.mediapipe_nFrames_nTrackedPoints_XYZ = np.load(str(self.mediapipe_3d_path))
            self.mediapipe_nFrames_nTrackedPoints_XYZ = self.mediapipe_nFrames_nTrackedPoints_XYZ/1000 #convert to meters
        else:
            # load charuco data        
            self.charuco_nFrames_nTrackedPoints_XYZ = fmc_session_obj.charuco_nFrames_nTrackedPoints_XYZ/1000 #convert to meters

            # load mediapipe data
            self.mediapipe_nFrames_nTrackedPoints_XYZ = fmc_session_obj.mediapipe_nFrames_nTrackedPoints_XYZ/1000 #convert to meters

        self.frame_num = 0
        self.num_frames = self.charuco_nFrames_nTrackedPoints_XYZ.shape[0]


        self.charuco_gl_scatter_item = gl.GLScatterPlotItem(pos=self.charuco_nFrames_nTrackedPoints_XYZ[self.frame_num,:,:], color=(1,1,1,.3), size=.1, pxMode=False)
        self.gl_view_widget.addItem(self.charuco_gl_scatter_item)
        
        self.charuco_gl_lineplot_item = gl.GLLinePlotItem(pos=self.charuco_nFrames_nTrackedPoints_XYZ[self.frame_num,:,:], color=(0,1,1,1), width=1., antialias=True)
        self.gl_view_widget.addItem(self.charuco_gl_lineplot_item)


        # Plot Media Pipe stuff
        self.mediapipe_gl_scatter_item = gl.GLScatterPlotItem(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num,:,:], color=(1,0,1,.3), size=.025, pxMode=False)
        self.gl_view_widget.addItem(self.mediapipe_gl_scatter_item)

        #mediapipe body connect-the-dots order
        self.mediapipe_head_indexes = [8, 6, 5, 4, 0, 10, 9, 0, 1, 2, 3, 7 ]
        self.mediapipe_tors_indexes = [12, 11, 24, 23, 12]
        self.mediapipe_rArm_indexes = [12, 14, 16, 18, 20, 16, 22 ]
        self.mediapipe_lArm_indexes = [11, 13, 15, 17, 19, 15, 21]
        self.mediapipe_rLeg_indexes = [24, 26, 28, 30, 32, 28, ]
        self.mediapipe_lLeg_indexes = [23, 25, 27, 29, 31, 27, ]
        
        rHandIDstart = 33
        lHandIDstart = 54
        mediapipe_hand_indexes = [0,1,2,3,4,3,2,1,0,5,6,7,8,7,6,5,9,10,11,12,11,10,9,13,14,15,16,15,14,13,17,18,19,20,19,18,17,0]
        
        
        self.mediapipe_rHand_indexes = [this_index + rHandIDstart for this_index in mediapipe_hand_indexes]
        self.mediapipe_lHand_indexes = [this_index + lHandIDstart for this_index in mediapipe_hand_indexes]
        
        self.mediapipe_head_gl_lineplot_item = gl.GLLinePlotItem(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num,self.mediapipe_head_indexes,:], color=(1,0,1,1), width=1., antialias=True)                
        self.mediapipe_tors_gl_lineplot_item = gl.GLLinePlotItem(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num,self.mediapipe_tors_indexes,:], color=(1,0,1,1), width=1., antialias=True)                        
        self.mediapipe_rArm_gl_lineplot_item = gl.GLLinePlotItem(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num,self.mediapipe_rArm_indexes,:], color=(1,0,1,1), width=1., antialias=True)                
        self.mediapipe_lArm_gl_lineplot_item = gl.GLLinePlotItem(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num,self.mediapipe_lArm_indexes,:], color=(1,0,1,1), width=1., antialias=True)                
        self.mediapipe_rLeg_gl_lineplot_item = gl.GLLinePlotItem(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num,self.mediapipe_rLeg_indexes,:], color=(1,0,1,1), width=1., antialias=True)                
        self.mediapipe_lLeg_gl_lineplot_item = gl.GLLinePlotItem(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num,self.mediapipe_lLeg_indexes,:], color=(1,0,1,1), width=1., antialias=True)
        
        self.mediapipe_rHand_gl_lineplot_item = gl.GLLinePlotItem(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num,self.mediapipe_rHand_indexes,:], color=(1,1,0,1), width=2., antialias=True)        
        self.mediapipe_lHand_gl_lineplot_item = gl.GLLinePlotItem(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num,self.mediapipe_lHand_indexes,:], color=(0,1,1,1), width=2., antialias=True)
        
        
        self.gl_view_widget.addItem(self.mediapipe_head_gl_lineplot_item)
        self.gl_view_widget.addItem(self.mediapipe_tors_gl_lineplot_item)
        self.gl_view_widget.addItem(self.mediapipe_rArm_gl_lineplot_item)
        self.gl_view_widget.addItem(self.mediapipe_lArm_gl_lineplot_item)
        self.gl_view_widget.addItem(self.mediapipe_rLeg_gl_lineplot_item)
        self.gl_view_widget.addItem(self.mediapipe_lLeg_gl_lineplot_item)
        self.gl_view_widget.addItem(self.mediapipe_rHand_gl_lineplot_item)
        self.gl_view_widget.addItem(self.mediapipe_lHand_gl_lineplot_item)


        self.qt_timer = QtCore.QTimer()
        self.qt_timer.timeout.connect(self.update)
        self.qt_timer.start(50)



    def update(self):
        
        self.frame_num +=1 
        if self.frame_num > self.num_frames:
            self.frame_num = 0
        
            
        self.charuco_gl_scatter_item.setData(pos=self.charuco_nFrames_nTrackedPoints_XYZ[self.frame_num,:,:])
        self.charuco_gl_lineplot_item.setData(pos=self.charuco_nFrames_nTrackedPoints_XYZ[self.frame_num,:,:])
        
        self.mediapipe_gl_scatter_item.setData(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num,:,:])
        
        self.mediapipe_head_gl_lineplot_item.setData(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num, self.mediapipe_head_indexes,:])
        self.mediapipe_tors_gl_lineplot_item.setData(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num, self.mediapipe_tors_indexes,:])
        self.mediapipe_rArm_gl_lineplot_item.setData(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num, self.mediapipe_rArm_indexes,:])
        self.mediapipe_lArm_gl_lineplot_item.setData(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num, self.mediapipe_lArm_indexes,:])
        self.mediapipe_rLeg_gl_lineplot_item.setData(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num, self.mediapipe_rLeg_indexes,:])
        self.mediapipe_lLeg_gl_lineplot_item.setData(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num, self.mediapipe_lLeg_indexes,:])

        self.mediapipe_rHand_gl_lineplot_item.setData(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num, self.mediapipe_rHand_indexes,:])
        self.mediapipe_lHand_gl_lineplot_item.setData(pos=self.mediapipe_nFrames_nTrackedPoints_XYZ[self.frame_num, self.mediapipe_lHand_indexes,:])

        if self.frame_num % 10 == 0:
            print("Frame Num : " + str(self.frame_num) + " - ViewPoint - Distance: " + str(self.gl_view_widget.opts['distance']) + " Elevation: " + str(self.gl_view_widget.opts['elevation']) + " Azimuth: " + str(self.gl_view_widget.opts['azimuth']) )
    
    def start(self):
        if (sys.flags.interactive != 1) or not hasattr(QtCore, "PYQT_VERSION"):
            QtGui.QApplication.instance().exec_()        


if __name__ == '__main__':
        visualizer = FMC_Visualizer(recording_session_path = "C:/Users/jonma/Dropbox/FreeMoCapProject/FreeMocap_Data/FreeMoCap_Session_2021-12-09_11_05_31")
        visualizer.start()
        



#%%
from pathlib import Path
import freemocap as fmc
import datetime
import logging
from rich import inspect
from rich import print
from rich.console import Console
console = Console()
import shutil
import pandas as pd
#%%



import socket
this_computer_name = socket.gethostname()

freemocap_data_path=None

if this_computer_name=='jon-hallway-XPS-8930':
    freemocap_data_path = Path('/home/jon/Dropbox/FreeMoCapProject/FreeMocap_Data')

elif this_computer_name == 'DESKTOP-DCG6K4F':
    freemocap_data_path = Path(r'C:\Users\jonma\Dropbox\FreeMoCapProject\FreeMocap_Data')
elif this_computer_name == 'DESKTOP-V3D343U':
    freemocap_data_path = Path(r'C:\Users\WindowsPC_Hallway\Dropbox\FreeMoCapProject\FreeMocap_Data')
elif this_computer_name == 'Jons-MacBook-Pro.local':
    freemocap_data_path = Path('/Users/jon/Dropbox/FreeMoCapProject\FreeMocap_Data')

#%%

batch_process_ID = datetime.datetime.now().strftime("batch_process_%Y-%m-%d_%H_%M_%S")
path_to_log_file = freemocap_data_path / (batch_process_ID+'_log.txt')

logging.basicConfig(filename=str(path_to_log_file), filemode='w', format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', level=logging.INFO)
logging.info('Log file create :D')

#%%
out_vid_path = freemocap_data_path / 'output_animation_videos'
out_vid_path.mkdir(exist_ok=True)

csv_path = out_vid_path / 'viable_folders_status.csv'

detect_folders_to_process_bool = True
if detect_folders_to_process_bool:
    dataframe_init_dict = {'viable_folder_paths':[],
                           'outVid_exists_bool':[]}
    viable_folders_dataframe = pd.DataFrame(dataframe_init_dict)

    with console.status('Looking for viable folders (that contain synced vids)'):
        for this_folder in freemocap_data_path.glob('**/*'):
            if this_folder.is_dir():
                #check if there is a synced video folder in here
                for this_sub_folder in this_folder.iterdir():
                    if this_sub_folder.is_dir():
                        if this_sub_folder.name.lower()[0:4]=='sync':
                            logging.info('Viable Folder found at {}'.format(str(this_folder)))
                            out_vid_exists = False
                            for this_mp4 in this_folder.glob('*.mp4'):
                                out_vid_exists = True
                                # shutil.copy(this_mp4, str(out_vid_path))
                                # logging.info('OutVid - {} - Copied to - {}'.format(str(this_mp4), str(out_vid_path)))
                            viable_folders_dataframe.loc[len(viable_folders_dataframe.index)] = [str(this_folder), out_vid_exists]
    viable_folders_dataframe.to_csv(csv_path, index=False)
    logging.info('List of all viable session folders saved to {}'.format(str(csv_path)))
else: 
    try:
        viable_folders_dataframe = pd.read_csv(csv_path)
        logging.info('Loaded csv with list viable session folders from {}'.format(str(csv_path)))
    except Exception:
        console.print_exception()
        logging.error('Failed to load csv with list viable session folders from {}'.format(str(csv_path)))


# %%


# %%
reprocess_bool = False
if reprocess_bool:
    folders_to_process = viable_folders_dataframe[viable_folders_dataframe["outVid_exists_bool"]==False]
    for this_item in folders_to_process['viable_folder_paths']:
        this_session_id = Path(this_item).name
        print('Starting to process SessionID: {}'.format(this_session_id))
        logging.info('Starting to process SessionID: {}'.format(this_session_id))
        try:
            fmc.RunMe(sessionID = this_session_id, stage=3, useBlender=True, showAnimation=False)
            logging.info('SessionID: {} processed successfully!'.format(this_session_id))
        except Exception:
            console.print_exception(Exception)
            logging.info('SessionID: {} failed to process'.format(this_session_id))




from fastapi import APIRouter

from src.api.services.user_config import UserConfigService
from src.cameras.detection.cam_singleton import get_or_create_cams
from src.config.webcam_config import WebcamConfig

camera_router = APIRouter()


@camera_router.post("/camera/config")
async def config_cam(webcam_config: WebcamConfig):
    s = UserConfigService()
    s.set_webcam_config(webcam_config)
    return s.webcam_configs


@camera_router.get("/camera/detect")
async def get_cameras():
    return get_or_create_cams()



from dataclasses import dataclass
from pathlib import Path


@dataclass
class SaveOptions():
    path_to_save_video: Path
    camera_name: str
    frame_width: int
    frame_height: int
    frames_per_second: float
    fourcc: str = "MP4V"

    def __post_init__(self):
        self.video_filename = self.camera_name + "_synchronized_raw.mp4"
        path_to_save_video = None

    @property
    def full_path(self):
        return Path().joinpath(self.path_to_save_video, self.video_filename)



import threading
import cv2
import imutils
import os
import platform
import time 

import mediapipe as mp
import numpy as np


mp_drawing = mp.solutions.drawing_utils
mp_drawing_styles = mp.solutions.drawing_styles
mp_holistic = mp.solutions.holistic

class VideoSetup(threading.Thread):
    """
    Class to run and thread webcams for preview purposes
    """
    def __init__(self, camID, parameterDictionary, rotNum):
        self.camID = camID
        self.parameterDictionary = parameterDictionary
        self.rotNum = rotNum
        threading.Thread.__init__(self)

    def run(self):
        # print("Starting " + self.previewName)
        self.record(self.parameterDictionary, self.rotNum)

    def record(self, parameterDictionary, rotNum):
        exposure = parameterDictionary.get("exposure")
        resWidth = parameterDictionary.get("resWidth")
        resHeight = parameterDictionary.get("resHeight")
        camWindowName = "Camera" + str(self.camID)+' Preview - Press ESC to exit Setup'
        cv2.namedWindow(camWindowName)

        if platform.system() == 'Windows':
            cap = cv2.VideoCapture(self.camID, cv2.CAP_DSHOW)
        else:
            cap = cv2.VideoCapture(self.camID, cv2.CAP_ANY)


        cap.set(cv2.CAP_PROP_FRAME_WIDTH, resWidth)
        cap.set(cv2.CAP_PROP_FRAME_HEIGHT, resHeight)
        cap.set(cv2.CAP_PROP_EXPOSURE, exposure)
        cap.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc('M', 'J', 'P', 'G')) 

        # showing values of the properties
        print("__________________________________________")
        print("cv2::videocapture properties for Camera# {}".format(self.camID))
        print("CV_CAP_PROP_FRAME_WIDTH: '{}'".format(cap.get(cv2.CAP_PROP_FRAME_WIDTH)))
        print("CV_CAP_PROP_FRAME_HEIGHT : '{}'".format(cap.get(cv2.CAP_PROP_FRAME_HEIGHT)))
        print("CAP_PROP_FPS : '{}'".format(cap.get(cv2.CAP_PROP_FPS)))
        print("CAP_PROP_EXPOSURE : '{}'".format(cap.get(cv2.CAP_PROP_EXPOSURE)))
        print("CAP_PROP_POS_MSEC : '{}'".format(cap.get(cv2.CAP_PROP_POS_MSEC)))
        print("CAP_PROP_FRAME_COUNT  : '{}'".format(cap.get(cv2.CAP_PROP_FRAME_COUNT)))
        print("CAP_PROP_BRIGHTNESS : '{}'".format(cap.get(cv2.CAP_PROP_BRIGHTNESS)))
        print("CAP_PROP_CONTRAST : '{}'".format(cap.get(cv2.CAP_PROP_CONTRAST)))
        print("CAP_PROP_SATURATION : '{}'".format(cap.get(cv2.CAP_PROP_SATURATION)))
        print("CAP_PROP_HUE : '{}'".format(cap.get(cv2.CAP_PROP_HUE)))
        print("CAP_PROP_GAIN  : '{}'".format(cap.get(cv2.CAP_PROP_GAIN)))
        print("CAP_PROP_CONVERT_RGB : '{}'".format(cap.get(cv2.CAP_PROP_CONVERT_RGB)))
        print("__________________________________________")

        while True:
            ret1, frame1 = cap.read()
            charuco_corners, charuco_ids, aruco_square_corners, aruco_square_ids = detect_charuco_board(frame1)
            annotate_image_with_charuco_data(frame1, charuco_corners, charuco_ids)
            
            if ret1 == True:
                if rotNum is not None:
                    frame1 = imutils.rotate_bound(frame1, angle=rotNum)
                
                cv2.imshow(camWindowName, frame1)
                if cv2.waitKey(1) & 0xFF == 27:
                    # == ord('q') for q
                    break

            else:
                break
        cv2.destroyWindow(camWindowName)

class MediaPipeVideoSetup(threading.Thread):
    """
    Class to run and thread webcams for preview purposes
    """
    def __init__(self, camID, parameterDictionary, rotNum):
        self.camID = camID
        self.parameterDictionary = parameterDictionary
        self.rotNum = rotNum
        threading.Thread.__init__(self)

    def run(self):
        # print("Starting " + self.previewName)
        self.record(self.parameterDictionary, self.rotNum)

    def record(self, parameterDictionary, rotNum):
        exposure = parameterDictionary.get("exposure")
        resWidth = parameterDictionary.get("resWidth")
        resHeight = parameterDictionary.get("resHeight")
        camWindowName = "Camera" + str(self.camID)+' - Press ESC to exit'

        cv2.namedWindow(camWindowName)

        if platform.system() == 'Windows':
            cap = cv2.VideoCapture(self.camID, cv2.CAP_DSHOW)
        else:
            cap = cv2.VideoCapture(self.camID, cv2.CAP_ANY)

        with mp_holistic.Holistic(
            static_image_mode=False,
            model_complexity=0,
            enable_segmentation=True) as holistic:

            cap.set(cv2.CAP_PROP_FRAME_WIDTH, resWidth)
            cap.set(cv2.CAP_PROP_FRAME_HEIGHT, resHeight)
            cap.set(cv2.CAP_PROP_EXPOSURE, exposure)
            cap.set(cv2.CAP_PROP_FOURCC, cv2.VideoWriter_fourcc('M', 'J', 'P', 'G')) 

                # showing values of the properties
            print("__________________________________________")
            print("cv2::videocapture properties for Camera# {}".format(self.camID))
            print("CV_CAP_PROP_FRAME_WIDTH: '{}'".format(cap.get(cv2.CAP_PROP_FRAME_WIDTH)))
            print("CV_CAP_PROP_FRAME_HEIGHT : '{}'".format(cap.get(cv2.CAP_PROP_FRAME_HEIGHT)))
            print("CAP_PROP_FPS : '{}'".format(cap.get(cv2.CAP_PROP_FPS)))
            print("CAP_PROP_EXPOSURE : '{}'".format(cap.get(cv2.CAP_PROP_EXPOSURE)))
            print("CAP_PROP_POS_MSEC : '{}'".format(cap.get(cv2.CAP_PROP_POS_MSEC)))
            print("CAP_PROP_FRAME_COUNT  : '{}'".format(cap.get(cv2.CAP_PROP_FRAME_COUNT)))
            print("CAP_PROP_BRIGHTNESS : '{}'".format(cap.get(cv2.CAP_PROP_BRIGHTNESS)))
            print("CAP_PROP_CONTRAST : '{}'".format(cap.get(cv2.CAP_PROP_CONTRAST)))
            print("CAP_PROP_SATURATION : '{}'".format(cap.get(cv2.CAP_PROP_SATURATION)))
            print("CAP_PROP_HUE : '{}'".format(cap.get(cv2.CAP_PROP_HUE)))
            print("CAP_PROP_GAIN  : '{}'".format(cap.get(cv2.CAP_PROP_GAIN)))
            print("CAP_PROP_CONVERT_RGB : '{}'".format(cap.get(cv2.CAP_PROP_CONVERT_RGB)))
            print("__________________________________________")
            timestamps = []
            while True:
                ret1, frame1 = cap.read()
                timestamps.append(time.time())
                
                #print(f"mean fps for cam {self.camID} is {1/np.mean(np.diff(timestamps))}")
                
                if ret1 == True:

                    charuco_corners, charuco_ids, aruco_square_corners, aruco_square_ids = detect_charuco_board(frame1)
                    try:
                        results = holistic.process(cv2.cvtColor(frame1, cv2.COLOR_BGR2RGB))
                    except Exception as e:
                        print(e)
                
                    annotate_image_with_charuco_data(frame1, charuco_corners, charuco_ids)
        
                    frame1.flags.writeable = False
                    try:
                        results = holistic.process(cv2.cvtColor(frame1, cv2.COLOR_BGR2RGB))
                    except Exception as e:
                        print(e)
                    mp_drawing.draw_landmarks(
                        frame1,
                        results.face_landmarks,
                        mp_holistic.FACEMESH_CONTOURS,
                        landmark_drawing_spec=None,
                        connection_drawing_spec=mp_drawing_styles
                        .get_default_face_mesh_contours_style())
                    mp_drawing.draw_landmarks(
                        frame1,
                        results.pose_landmarks,
                        mp_holistic.POSE_CONNECTIONS,
                        landmark_drawing_spec=mp_drawing_styles
                        .get_default_pose_landmarks_style())
                    mp_drawing.draw_landmarks(
                        frame1,
                        results.left_hand_landmarks,
                        mp_holistic.HAND_CONNECTIONS,
                        landmark_drawing_spec=mp_drawing_styles
                        .get_default_hand_landmarks_style())
                    mp_drawing.draw_landmarks(
                        frame1,
                        results.right_hand_landmarks,
                        mp_holistic.HAND_CONNECTIONS,
                        landmark_drawing_spec=mp_drawing_styles
                        .get_default_hand_landmarks_style())
                    if rotNum is not None:
                        frame1 = imutils.rotate_bound(frame1, angle=rotNum)
                        
                        
                    
                    
                    cv2.imshow(camWindowName, frame1)
                    if cv2.waitKey(1) & 0xFF == 27:
                        # == ord('q') for q
                        break

                else:
                    break
            cv2.destroyWindow(camWindowName)



def RunSetup(cam_inputs, rotation_input, paramDict,mediaPipeOverlay):
    """
    Start video setup by threading instances of the VideoSetup class
    """
    if not cam_inputs:
        raise ValueError("Camera input list (cam_inputs) is empty")

    ulist = []

    for cam_input, cam_rotation in zip(cam_inputs, rotation_input):
        if mediaPipeOverlay == True:
            u = MediaPipeVideoSetup(cam_input, paramDict, cam_rotation)
        else:
            u = VideoSetup(cam_input, paramDict, cam_rotation)
        u.start()
        ulist.append(u)

    for k in ulist:
        k.join()





def detect_charuco_board(image,  annotate_image=True):
    """
    Charuco base pose estimation.
    more-or-less copied from - https://mecaruco2.readthedocs.io/en/latest/notebooks_rst/Aruco
    /sandbox/ludovic/aruco_calibration_rotation.html
    """
    charuco_corners = []
    charuco_ids = []
    
    aruco_dict = cv2.aruco.Dictionary_get(cv2.aruco.DICT_4X4_250)
    charuco_length = 7
    charuco_width = 5

    board = cv2.aruco.CharucoBoard_create(charuco_length, charuco_width, 1, .8, aruco_dict)
    global num_charuco_corners
    num_charuco_corners = (charuco_length-1) * (charuco_width-1)


    # SUB PIXEL CORNER DETECTION CRITERION
    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 100, 0.00001)

    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    aruco_square_corners, aruco_square_ids, rejectedImgPoints = cv2.aruco.detectMarkers(gray,
        aruco_dict)

    if len(aruco_square_corners) > 0:
        # SUB PIXEL DETECTION
        for this_corner in aruco_square_corners:
            cv2.cornerSubPix(gray, this_corner,
                winSize=(3, 3),
                zeroZone=(-1, -1),
                criteria=criteria)
        res2 = cv2.aruco.interpolateCornersCharuco(aruco_square_corners, aruco_square_ids, gray,
            board)

        if res2[1] is not None and res2[2] is not None and len(res2[1]) > 3:
            charuco_corners = res2[1]
            charuco_ids = res2[2]


    return charuco_corners, charuco_ids, aruco_square_corners, aruco_square_ids

def annotate_image_with_charuco_data(image, charuco_corners, charuco_ids)->bool:

    full_charuco_detected_on_this_frame = False
    if len(charuco_ids) == num_charuco_corners:
        full_charuco_detected_on_this_frame = True

    image_w_markers = cv2.aruco.drawDetectedCornersCharuco(image,
                                                            np.array(charuco_corners),
                                                            np.array(charuco_ids),
                                                            (200,100,200,255)) #I think cv2 uses BGR instead of RGB?


    text_to_write_on_this_camera = ''
    current_cam_corner_count_str = str(
        len(charuco_ids)) + " of " + str(
        num_charuco_corners) + " ChAruco Corner Points detected | Full Board Detected: " + str(
        full_charuco_detected_on_this_frame)
    # TODO - Determine 'shared views' (i.e. frames in which a full board is detected by 2 cameras)
    # TODO - self.determine_shared_charuco_board_views()
    # this_cam_shared_views_str = " | Shared Views: " + str(
    #     each_cameras_shared_board_view_count_total)
    text_to_write_on_this_camera = current_cam_corner_count_str

    position = (10, 50)
    cv2.putText(
        image_w_markers,  # numpy array on which text is written
        text_to_write_on_this_camera,  # text
        position,  # position at which writing has to start
        cv2.FONT_HERSHEY_SIMPLEX,  # font family
        .5,  # font size
        (0, 10, 0, 255),  # font color
        4)  # font stroke (draw a darker heavier font beneath a lighter/thinner copy for readability)

    cv2.putText(
        image_w_markers,  # numpy array on which text is written
        text_to_write_on_this_camera,  # text
        position,  # position at which writing has to start
        cv2.FONT_HERSHEY_SIMPLEX,  # font family (very limited selection, i think there's some interesting CV history here...)
        .5,  # font size
        (209, 180, 0, 255),  # font color
        2 ) # font stroke

    if full_charuco_detected_on_this_frame:
        cv2.polylines(image_w_markers, np.int32([charuco_corners]), True, (0, 255, 255), 4)
        # for these_corners in charuco_points_from_previous_frames:
        # if len(these_corners)>0:
        #     cv2.polylines(image_w_markers, np.int32([these_corners]), True, (0,100,255,255/2), 2)

    return True


from dataclasses import dataclass
from typing import Dict

import cv2
import numpy as np


@dataclass
class CharucoBoard:
    _aruco_marker_dict: Dict = cv2.aruco.Dictionary_get(cv2.aruco.DICT_4X4_250)
    _number_of_squares_width: int = 7
    _number_of_squares_height: int = 5
    _black_square_side_length: int = 1
    _aruco_marker_length_proportional: float = .8

    def __post_init__(self):
        self._charuco_board = cv2.aruco.CharucoBoard_create(self._number_of_squares_width,
                                                            self._number_of_squares_height,
                                                            self._black_square_side_length,
                                                            self._aruco_marker_length_proportional,
                                                            self._aruco_marker_dict)

        self._number_of_charuco_corners = (self._number_of_squares_width - 1) * (self._number_of_squares_height - 1)


    @property
    def charuco_board(self):
        return self._charuco_board

    @property
    def number_of_charuco_corners(self):
        return self._number_of_charuco_corners

    @property
    def aruco_marker_dict(self):
        return self._aruco_marker_dict

